"""Sphinx configuration."""
from datetime import datetime

import gptsum

project = "gptsum"
author = "Nicolas Trangez"
copyright = f"{datetime.now().year}, {author}"  # noqa: A001

version = gptsum.__version__
release = version

html_theme = "furo"

extensions = [
    "sphinx.ext.autodoc",
    "sphinxarg.ext",
]



"""Test code from `README.rst`."""

import shutil
import subprocess  # noqa: S404
from pathlib import Path
from typing import Optional

import pytest


@pytest.mark.skipif(shutil.which("sfdisk") is None, reason="sfdisk not found")
def test_quickstart(tmp_path: Path) -> None:
    """Run the 'Quickstart' section from 'README.rst'."""

    def check_call(cmd: bytes) -> None:
        subprocess.check_call(
            args=[cmd],
            shell=True,
            cwd=tmp_path,
        )

    def check_output(cmd: bytes, input_: Optional[bytes] = None) -> bytes:
        return subprocess.check_output(
            args=[cmd],
            shell=True,
            cwd=tmp_path,
            input=input_ or b"",
            stderr=subprocess.STDOUT,
        )

    check_call(b"truncate -s64M image.raw")

    sfdisk_input = b"""
label: gpt
label-id: 132e3631-1ec9-4411-ab25-9b95b54b0903
first-lba: 2048
"""
    expected_sfdisk_output = b"""\
Checking that no-one is using this disk right now ... OK

Disk image.raw: 64 MiB, 67108864 bytes, 131072 sectors
Units: sectors of 1 * 512 = 512 bytes
Sector size (logical/physical): 512 bytes / 512 bytes
I/O size (minimum/optimal): 512 bytes / 512 bytes

>>> Script header accepted.
>>> Script header accepted.
>>> Script header accepted.
>>> Done.
Created a new GPT disklabel (GUID: 132E3631-1EC9-4411-AB25-9B95B54B0903).

New situation:
Disklabel type: gpt
Disk identifier: 132E3631-1EC9-4411-AB25-9B95B54B0903

The partition table has been altered.
Syncing disks.
"""
    assert (
        check_output(b"sfdisk image.raw", input_=sfdisk_input) == expected_sfdisk_output
    )

    assert (
        check_output(b"gptsum get-guid image.raw")
        == b"132e3631-1ec9-4411-ab25-9b95b54b0903\n"
    )

    assert check_output(b'gptsum verify image.raw || echo "Verification failed!"') == (
        b"Disk GUID doesn't match expected checksum, "
        b"got 132e3631-1ec9-4411-ab25-9b95b54b0903, "
        b"expected 6190f5bb-1967-14ec-9fbd-a7d213a45461\n"
        b"Verification failed!\n"
    )

    check_call(b"gptsum embed image.raw")

    assert (
        check_output(b'gptsum verify image.raw && echo "Verification succeeded!"')
        == b"Verification succeeded!\n"
    )

    assert (
        check_output(b"gptsum get-guid image.raw")
        == b"6190f5bb-1967-14ec-9fbd-a7d213a45461\n"
    )



"""Global definitions for pytest tests."""

import shutil
import tempfile
import uuid
from pathlib import Path
from typing import Iterator

import pytest

TESTDATA_DISK = Path(__file__).parent / "testdata" / "disk"
TESTDATA_DISK_GUID = uuid.UUID("66E0318D-A103-9549-8583-80E8ABCD4CD8")

TESTDATA_EMBEDDED_DISK = Path(__file__).parent / "testdata" / "embedded-disk"
TESTDATA_EMBEDDED_DISK_GUID = uuid.UUID("D4750646-01FD-2608-959F-159017007377")


@pytest.fixture()
def disk_image(tmp_path: Path) -> Iterator[Path]:
    """Yield the path to a copy of `TESTDATA_DISK`."""
    with tempfile.NamedTemporaryFile(dir=tmp_path) as tmp:
        with open(TESTDATA_DISK, "rb") as disk:
            shutil.copyfileobj(disk, tmp)

        yield Path(tmp.name)



"""Tests for :mod:`gptsum`."""

import hashlib
import os
import uuid
from pathlib import Path
from typing import Any, Callable, List

import pytest

import gptsum
from gptsum import checksum
from tests import conftest


@pytest.mark.parametrize(("attr"), ["author", "contact", "license", "version"])
def test_attribute(attr: str) -> None:
    """Test expected metadata attributes on the package."""
    real_attr = "__{}__".format(attr)
    assert hasattr(gptsum, real_attr)
    assert type(getattr(gptsum, real_attr)) is str


@pytest.mark.parametrize(
    ("func", "args"),
    [
        (gptsum.get_guid, []),
        (gptsum.set_guid, [uuid.UUID("656156c4-2456-4038-b7f3-dfa0ae263e95")]),
        (gptsum.calculate_expected_guid, []),
        (gptsum.embed, []),
        (gptsum.verify, []),
    ],
)
def test_argument_conflicts(func: Callable[..., Any], args: List[Any]) -> None:
    """Ensure ``fd``/``path`` can't be passed at the same time."""
    with pytest.raises(ValueError, match="Either fd or path must be given"):
        func(*args, fd=None, path=None)

    with pytest.raises(ValueError, match="Both fd and path can't be given"):
        func(*args, fd=0, path=Path("/"))


@pytest.mark.parametrize(
    ("disk_file", "expected_guid"),
    [
        (conftest.TESTDATA_DISK, conftest.TESTDATA_DISK_GUID),
        (conftest.TESTDATA_EMBEDDED_DISK, conftest.TESTDATA_EMBEDDED_DISK_GUID),
    ],
)
def test_get_guid(disk_file: Path, expected_guid: uuid.UUID) -> None:
    """Test :func:`gptsum.get_guid`."""
    assert gptsum.get_guid(path=disk_file) == expected_guid

    with open(disk_file, "rb") as fd:
        assert gptsum.get_guid(fd=fd.fileno()) == expected_guid


@pytest.mark.parametrize(("method"), ["by_fd", "by_path"])
def test_set_guid(method: str, disk_image: Path) -> None:
    """Test :func:`gptsum.set_guid`."""
    new_guid = uuid.UUID("e81f41f2-1807-49aa-876b-57a7e727b3f8")

    assert gptsum.get_guid(path=disk_image) != new_guid

    if method == "by_fd":
        with open(disk_image, "rb+") as fd:
            gptsum.set_guid(new_guid, fd=fd.fileno())
    else:
        assert method == "by_path"
        gptsum.set_guid(new_guid, path=disk_image)

    assert gptsum.get_guid(path=disk_image) == new_guid


@pytest.mark.parametrize(
    ("disk_file", "expected_guid"),
    [
        (conftest.TESTDATA_DISK, conftest.TESTDATA_EMBEDDED_DISK_GUID),
        (conftest.TESTDATA_EMBEDDED_DISK, conftest.TESTDATA_EMBEDDED_DISK_GUID),
    ],
)
def test_calculate_expected_guid(disk_file: Path, expected_guid: uuid.UUID) -> None:
    """Test :func:`gptsum.calculate_expected_guid`."""
    assert gptsum.calculate_expected_guid(path=disk_file) == expected_guid

    with open(disk_file, "rb") as fd:
        assert gptsum.calculate_expected_guid(fd=fd.fileno()) == expected_guid


@pytest.mark.parametrize(("method"), ["by_fd", "by_path"])
def test_embed(method: str, disk_image: Path) -> None:
    """Test :func:`gptsum.embed`."""
    if method == "by_fd":
        with open(disk_image, "rb+") as fd:
            gptsum.embed(fd=fd.fileno())
    else:
        assert method == "by_path"
        gptsum.embed(path=disk_image)

    guid = gptsum.get_guid(path=disk_image)
    assert guid == conftest.TESTDATA_EMBEDDED_DISK_GUID

    hash1 = hashlib.sha256()
    with open(conftest.TESTDATA_EMBEDDED_DISK, "rb") as fd:
        size = os.fstat(fd.fileno()).st_size
        done = checksum.hash_file(hash1.update, fd.fileno(), size, 0)
        assert done == size

    hash2 = hashlib.sha256()
    with open(disk_image, "rb") as fd:
        size = os.fstat(fd.fileno()).st_size
        done = checksum.hash_file(hash2.update, fd.fileno(), size, 0)
        assert done == size

    assert hash1.hexdigest() == hash2.hexdigest()


def test_embed_noop(disk_image: Path) -> None:
    """Test to ensure issuing :option:`embed` on an already-prepared file is a no-op."""
    gptsum.embed(path=disk_image)
    stat1 = os.stat(disk_image)
    gptsum.embed(path=disk_image)
    stat2 = os.stat(disk_image)

    assert stat2.st_mtime == stat1.st_mtime
    assert stat2.st_ctime == stat1.st_ctime


@pytest.mark.parametrize(("method"), ["by_fd", "by_path"])
def test_verify(method: str, disk_image: Path) -> None:
    """Test :func:`gptsum.verify`."""
    if method == "by_fd":
        with open(disk_image, "rb") as fd:
            with pytest.raises(gptsum.VerificationFailure):
                gptsum.verify(fd=fd.fileno())
    else:
        assert method == "by_path"
        with pytest.raises(gptsum.VerificationFailure):
            gptsum.verify(path=disk_image)

    gptsum.embed(path=disk_image)

    if method == "by_fd":
        with open(disk_image, "rb") as fd:
            gptsum.verify(fd=fd.fileno())
    else:
        assert method == "by_path"
        gptsum.verify(path=disk_image)



"""Tests for the :mod:`gptsum.checksum` module."""

import hashlib
import os
from pathlib import Path
from typing import BinaryIO

import pytest_benchmark.fixture  # type: ignore[import]

from gptsum import checksum, gpt
from tests import conftest


def blake2b(fd: BinaryIO) -> bytes:
    """Calculate the Blake2b digest of a file.

    :param fd: File-like object of which to calculate the digest

    :returns: Blake2b digest of the file contents
    """
    hasher = hashlib.blake2b(digest_size=16)

    size = os.fstat(fd.fileno()).st_size
    checksum.hash_file(hasher.update, fd.fileno(), size, 0)

    return hasher.digest()


def test_calculate() -> None:
    """Test checksum calculation of an image twice."""
    with gpt.GPTImage(path=conftest.TESTDATA_DISK, open_mode=os.O_RDONLY) as image:
        digest1 = checksum.calculate(image)
        digest2 = checksum.calculate(image)

        assert digest2 == digest1
        assert digest1 == conftest.TESTDATA_EMBEDDED_DISK_GUID.bytes

    with gpt.GPTImage(
        path=conftest.TESTDATA_EMBEDDED_DISK, open_mode=os.O_RDONLY
    ) as image:
        assert checksum.calculate(image) == conftest.TESTDATA_EMBEDDED_DISK_GUID.bytes


def test_calculate_inplace(disk_image: Path) -> None:
    """Test :func:`checksum.calculate` by modifying in image in-place."""
    with open(disk_image, "rb") as fd:
        real_hash = blake2b(fd)

    with gpt.GPTImage(path=disk_image, open_mode=os.O_RDONLY) as image:
        digest = checksum.calculate(image)
        assert digest != real_hash

    # Overwrite CRCs and GUIDs with zeros, in-place
    with open(disk_image, "rb+") as fd:
        header_crc32_offset = 8 + 4 + 4
        guid_offset = header_crc32_offset + 4 + 4 + 8 + 8 + 8 + 8

        gpt.pwrite_all(fd.fileno(), b"\0" * 4, gpt.MBR_SIZE + header_crc32_offset)
        gpt.pwrite_all(fd.fileno(), b"\0" * 16, gpt.MBR_SIZE + guid_offset)

        size = os.fstat(fd.fileno()).st_size
        gpt.pwrite_all(
            fd.fileno(), b"\0" * 4, size - gpt.LBA_SIZE + header_crc32_offset
        )
        gpt.pwrite_all(fd.fileno(), b"\0" * 16, size - gpt.LBA_SIZE + guid_offset)

    with open(disk_image, "rb") as fd:
        new_hash = blake2b(fd)

        assert new_hash == digest


def test_calculate_benchmark(
    benchmark: pytest_benchmark.fixture.BenchmarkFixture,
) -> None:
    """Benchmark :func:`checksum.calculate`."""
    with gpt.GPTImage(path=conftest.TESTDATA_DISK, open_mode=os.O_RDONLY) as image:
        benchmark(checksum.calculate, image)


def test_digest_to_guid() -> None:
    """Test :func:`checksum.digest_to_guid`."""
    hasher = hashlib.blake2b(digest_size=16)
    hasher.update(b"\0" * 32)
    digest = hasher.digest()
    guid = checksum.digest_to_guid(digest)
    assert guid.bytes == digest



"""Tests for the :mod:`gptsum.gpt` module."""

import os
import struct
import tempfile
import uuid
from pathlib import Path
from typing import Iterator, Type

import pytest

from gptsum import gpt
from tests import conftest


@pytest.mark.parametrize(
    "header",
    [
        b"",
        b"\0",
        b"\0" * (gpt.LBA_SIZE - 1),
        b"\0" * (gpt.LBA_SIZE + 1),
    ],
)
def test_gptheader_unpack_invalid_header_size(header: bytes) -> None:
    """Test unpacking GPT headers of invalid size."""
    with pytest.raises(gpt.InvalidHeaderSizeError):
        gpt.GPTHeader.unpack(header)


# Generated using:
# $ truncate -s1G disk
# $ sfdisk disk << EOF
# label: gpt
# label-id: 97fe4bef-3385-45d8-becf-e0e178e0d1d9
# first-lba: 2048
# EOF
# Then, peel out bytes [512, 1024[ and [1024 * 1024 * 1024 - 512, 1024 * 1024 * 1024[
TEST_DISK_SIZE = 1024 * 1024 * 1024
TEST_DISK_UUID = uuid.UUID("97fe4bef-3385-45d8-becf-e0e178e0d1d9")
TEST_DISK_NUM_ENTRIES = 128
TEST_DISK_ENTRY_SIZE = 128
TEST_DISK_GPT_HEADER = (
    b"EFI PART\x00\x00\x01\x00\\\x00\x00\x00\xf9\xc2\x1d\x95\x00\x00\x00\x00\x01\x00"
    b"\x00\x00\x00\x00\x00\x00\xff\xff\x1f\x00\x00\x00\x00\x00\x00\x08\x00\x00\x00"
    b"\x00\x00\x00\xde\xff\x1f\x00\x00\x00\x00\x00\xefK\xfe\x97\x853\xd8E\xbe\xcf\xe0"
    b"\xe1x\xe0\xd1\xd9\x02\x00\x00\x00\x00\x00\x00\x00\x80\x00\x00\x00\x80\x00\x00"
    b"\x00\x86\xd2T\xab\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00"
)
TEST_DISK_BACKUP_GPT_HEADER = (
    b"EFI PART\x00\x00\x01\x00\\\x00\x00\x00%\xb3\xe7\r\x00\x00\x00\x00\xff\xff\x1f"
    b"\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x08\x00\x00\x00\x00"
    b"\x00\x00\xde\xff\x1f\x00\x00\x00\x00\x00\xefK\xfe\x97\x853\xd8E\xbe\xcf\xe0\xe1x"
    b"\xe0\xd1\xd9\xdf\xff\x1f\x00\x00\x00\x00\x00\x80\x00\x00\x00\x80\x00\x00\x00\x86"
    b"\xd2T\xab\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00"
)


@pytest.mark.parametrize(
    ("raw_header", "current_lba", "backup_lba", "entries_starting_lba"),
    [
        (TEST_DISK_GPT_HEADER, 1, TEST_DISK_SIZE // 512 - 1, 2),
        (
            TEST_DISK_BACKUP_GPT_HEADER,
            TEST_DISK_SIZE // 512 - 1,
            1,
            TEST_DISK_SIZE // 512
            - 1
            - (TEST_DISK_NUM_ENTRIES * TEST_DISK_ENTRY_SIZE // 512),
        ),
    ],
)
def test_gptheader_unpack(
    raw_header: bytes, current_lba: int, backup_lba: int, entries_starting_lba: int
) -> None:
    """Test unpacking of a valid GPT header."""
    header = gpt.GPTHeader.unpack(raw_header)

    assert header.current_lba == current_lba
    assert header.backup_lba == backup_lba
    assert header.first_usable_lba == 2048
    assert header.last_usable_lba == 2097118
    assert header.disk_guid == TEST_DISK_UUID
    assert header.entries_starting_lba == entries_starting_lba
    assert header.num_entries == TEST_DISK_NUM_ENTRIES
    assert header.entry_size == TEST_DISK_ENTRY_SIZE

    # Alternative, calculated way to find the right values
    assert (
        header.last_usable_lba
        == TEST_DISK_SIZE // 512
        - 1
        - (header.num_entries * header.entry_size // 512)
        - 1
    )


def _replace_signature(raw_header: bytes, signature: bytes) -> bytes:
    """Replace the 'signature' field in a raw header."""
    return signature + raw_header[8:]


def _replace_revision(raw_header: bytes, revision: bytes) -> bytes:
    """Replace the 'revision' field in a raw header."""
    return raw_header[:8] + revision + raw_header[8 + 4 :]


def _replace_header_size(raw_header: bytes, header_size: int) -> bytes:
    """Replace the 'header_size' field in a raw header."""
    header_size_bytes = struct.pack("<I", header_size)
    return raw_header[: 8 + 4] + header_size_bytes + raw_header[8 + 4 + 4 :]


def _replace_reserved(raw_header: bytes, reserved: bytes) -> bytes:
    """Replace the 'reserved' field in a raw header."""
    return raw_header[: 8 + 4 + 4 + 4] + reserved + raw_header[8 + 4 + 4 + 4 + 4 :]


def _replace_padding(raw_header: bytes, padding: bytes) -> bytes:
    """Replace the 'padding' field in a raw header."""
    return raw_header[:92] + padding


def _replace_header_crc32(raw_header: bytes, header_crc32: int) -> bytes:
    """Replace the 'header_crc32' field in a raw header."""
    header_crc32_bytes = struct.pack("<I", header_crc32)
    return raw_header[: 8 + 4 + 4] + header_crc32_bytes + raw_header[8 + 4 + 4 + 4 :]


@pytest.mark.parametrize(
    ("expected_error", "raw_header"),
    [
        (
            gpt.InvalidSignatureError,
            _replace_signature(TEST_DISK_GPT_HEADER, b"FFI PART"),
        ),
        (
            gpt.UnsupportedRevisionError,
            _replace_revision(TEST_DISK_GPT_HEADER, b"\0\0\1\1"),
        ),
        (gpt.InvalidHeaderSizeError, _replace_header_size(TEST_DISK_GPT_HEADER, 90)),
        (gpt.InvalidFieldError, _replace_reserved(TEST_DISK_GPT_HEADER, b"\0\0\0\1")),
        (
            gpt.InvalidFieldError,
            _replace_padding(TEST_DISK_GPT_HEADER, b"\1" * (512 - 92)),
        ),
        (
            gpt.HeaderChecksumMismatchError,
            _replace_header_crc32(TEST_DISK_GPT_HEADER, 2 ** 32 - 1),
        ),
    ],
)
def test_gptheader_unpack_error(
    expected_error: Type[Exception], raw_header: bytes
) -> None:
    """Test :meth:`gpt.GPTHeader.unpack` failure cases."""
    with pytest.raises(expected_error):
        gpt.GPTHeader.unpack(raw_header)


def test_gptheader_is_backup_of() -> None:
    """Test :meth:`gpt.GPTHeader.is_backup_of`."""
    gpt1 = gpt.GPTHeader.unpack(TEST_DISK_GPT_HEADER)
    gpt2 = gpt.GPTHeader.unpack(TEST_DISK_BACKUP_GPT_HEADER)

    assert gpt1.is_backup_of(gpt2)
    assert gpt2.is_backup_of(gpt1)

    assert not gpt1.is_backup_of(gpt1)
    assert not gpt2.is_backup_of(gpt2)


@pytest.mark.parametrize(
    "raw_header",
    [
        TEST_DISK_GPT_HEADER,
        TEST_DISK_BACKUP_GPT_HEADER,
    ],
)
def test_gptheader_pack(raw_header: bytes) -> None:
    """Test :meth:`gpt.GPTHeader.pack`."""
    assert gpt.GPTHeader.unpack(raw_header).pack() == raw_header


def test_gptheader_pack_override_crc32() -> None:
    """Test :meth:`gpt.GPTHeader.pack` overriding the CRC32."""
    header = gpt.GPTHeader.unpack(TEST_DISK_GPT_HEADER)
    packed = header.pack(override_crc32=0)
    with pytest.raises(gpt.HeaderChecksumMismatchError):
        gpt.GPTHeader.unpack(packed)


@pytest.fixture()
def small_file(tmp_path: Path) -> Iterator[Path]:
    """Yield the path of an empty, 1kB temporary file."""
    with tempfile.NamedTemporaryFile(dir=tmp_path) as tmp:
        tmp.truncate(1024)
        yield Path(tmp.name)


def test_gptimage_constructor(small_file: Path) -> None:
    """Test the :class:`gpt.GPTImage` constructor."""
    with pytest.raises(ValueError, match="Either fd or path must be given"):
        gpt.GPTImage()

    with pytest.raises(ValueError, match="Both fd and path can't be given"):
        gpt.GPTImage(fd=0, path=small_file)


def test_gptimage_too_small(small_file: Path) -> None:
    """Test :class:`gpt.GPTImage` with a (too) small file."""
    with open(small_file, "rb") as fd:
        with pytest.raises(gpt.InvalidImageError):  # noqa: PT012
            with gpt.GPTImage(fd=fd.fileno()):
                pytest.fail("This code should not be reached")  # pragma: no cover

    with pytest.raises(gpt.InvalidImageError):  # noqa: PT012
        with gpt.GPTImage(path=small_file):
            pytest.fail("This code should not be reached")  # pragma: no cover


def test_gptimage_validate() -> None:
    """Test :meth:`gpt.GPTImage.validate`."""
    with gpt.GPTImage(path=conftest.TESTDATA_DISK, open_mode=os.O_RDONLY) as image:
        image.validate()

    with open(conftest.TESTDATA_DISK, "rb") as fd:
        with gpt.GPTImage(fd=fd.fileno()) as image:
            image.validate()


def test_gptimage_validate_invalid_image(disk_image: Path) -> None:
    """Test :meth:`gpt.GPTImage.validate` with an invalid image."""
    with gpt.GPTImage(path=disk_image, open_mode=os.O_RDONLY) as image:
        primary = image.read_primary_gpt_header()

    new_guid = uuid.UUID("21da705c-fec8-4857-8379-449d823a7155")
    new_primary = gpt.GPTHeader(
        primary.current_lba,
        primary.backup_lba,
        primary.first_usable_lba,
        primary.last_usable_lba,
        new_guid,
        primary.entries_starting_lba,
        primary.num_entries,
        primary.entry_size,
        primary.entries_crc32,
    )

    with open(disk_image, "rb+") as fd:
        gpt.pwrite_all(fd.fileno(), new_primary.pack(), gpt.MBR_SIZE)

    with pytest.raises(gpt.InvalidImageError):  # noqa: PT012
        with gpt.GPTImage(path=disk_image, open_mode=os.O_RDONLY) as image:
            image.validate()


def test_gptimage_read_primary_gpt_header() -> None:
    """Test :meth:`gpt.GPTImage.read_primary_gpt_header`."""
    with gpt.GPTImage(path=conftest.TESTDATA_DISK, open_mode=os.O_RDONLY) as image:
        header = image.read_primary_gpt_header()
        assert header.disk_guid == conftest.TESTDATA_DISK_GUID
        assert header.first_usable_lba == 2048
        assert header.last_usable_lba == 4062
        assert header.current_lba == 1
        assert header.backup_lba == 4095


def test_gptimage_read_backup_gpt_header() -> None:
    """Test :meth:`gpt.GPTImage.read_backup_gpt_header`."""
    with gpt.GPTImage(path=conftest.TESTDATA_DISK, open_mode=os.O_RDONLY) as image:
        header = image.read_backup_gpt_header()
        assert header.disk_guid == conftest.TESTDATA_DISK_GUID
        assert header.first_usable_lba == 2048
        assert header.last_usable_lba == 4062
        assert header.current_lba == 4095
        assert header.backup_lba == 1


def test_gptimage_write_gpt_headers(disk_image: Path) -> None:
    """Test :meth:`gpt.GPTImage.write_gpt_headers`."""
    with gpt.GPTImage(path=disk_image, open_mode=os.O_RDONLY) as image:
        primary = image.read_primary_gpt_header()
        backup = image.read_backup_gpt_header()

    new_guid = uuid.UUID("086991f8-75bd-4560-8943-e11c0c59422b")

    new_primary = gpt.GPTHeader(
        primary.current_lba,
        primary.backup_lba,
        primary.first_usable_lba,
        primary.last_usable_lba,
        new_guid,
        primary.entries_starting_lba,
        primary.num_entries,
        primary.entry_size,
        primary.entries_crc32,
    )
    new_backup = gpt.GPTHeader(
        backup.current_lba,
        backup.backup_lba,
        backup.first_usable_lba,
        backup.last_usable_lba,
        new_guid,
        backup.entries_starting_lba,
        backup.num_entries,
        backup.entry_size,
        backup.entries_crc32,
    )

    with gpt.GPTImage(path=disk_image, open_mode=os.O_WRONLY) as image:
        image.write_gpt_headers(new_primary, new_backup)

    with gpt.GPTImage(path=disk_image, open_mode=os.O_RDONLY) as image:
        image.validate()
        assert image.read_primary_gpt_header().disk_guid == new_guid


def test_gptimage_write_gpt_headers_incompatible_headers(disk_image: Path) -> None:
    """Test :meth:`gpt.GPTImage.write_gpt_headers` with incompatible headers."""
    with gpt.GPTImage(path=disk_image, open_mode=os.O_RDONLY) as image:
        primary = image.read_primary_gpt_header()
        backup = image.read_backup_gpt_header()

        new_guid = uuid.UUID("913ef501-b10c-4fa4-8919-0d47f7d4d4bd")

        new_primary = gpt.GPTHeader(
            primary.current_lba,
            primary.backup_lba,
            primary.first_usable_lba,
            primary.last_usable_lba,
            new_guid,
            primary.entries_starting_lba,
            primary.num_entries,
            primary.entry_size,
            primary.entries_crc32,
        )

        with pytest.raises(
            ValueError, match="Given headers are not backups of each other"
        ):
            image.write_gpt_headers(new_primary, backup)


def test_gptimage_write_gpt_headers_incorrect_current_lba(disk_image: Path) -> None:
    """Test :meth:`gpt.GPTImage.write_gpt_headers` with incorrect headers."""
    with gpt.GPTImage(path=disk_image, open_mode=os.O_RDONLY) as image:
        primary = image.read_primary_gpt_header()
        backup = image.read_backup_gpt_header()

        with pytest.raises(
            ValueError, match="Primary header has invalid 'current_lba', expected 1"
        ):
            image.write_gpt_headers(backup, primary)

        with open(disk_image, "rb+") as fd:
            size = os.fstat(fd.fileno()).st_size
            fd.truncate(size + gpt.LBA_SIZE)

        with pytest.raises(
            ValueError, match=r"Backup header has invalid 'current_lba', expected \d+"
        ):
            image.write_gpt_headers(primary, backup)


def test_gptimage_update_guid(disk_image: Path) -> None:
    """Test :meth:`gpt.GPTImage.update_guid`."""
    new_guid = uuid.UUID("910fbf4b-f2ac-4d2c-8995-9dff42b9efde")

    with gpt.GPTImage(path=disk_image, open_mode=os.O_RDWR) as image:
        image.update_guid(new_guid)

    with gpt.GPTImage(path=disk_image, open_mode=os.O_RDONLY) as image:
        image.validate()
        assert image.read_primary_gpt_header().disk_guid == new_guid
        assert image.read_backup_gpt_header().disk_guid == new_guid



"""
gptsum: a tool to make disk images using GPT partitions self-verifiable.

Like `isomd5sum <https://github.com/rhinstaller/isomd5sum>`_.

This module exposes the functionality provided by the tool through a Python API.
"""

import dataclasses
import os
import sys
import uuid
from pathlib import Path
from typing import Callable, Optional, cast

if sys.version_info >= (3, 8):
    from importlib.metadata import Distribution, distribution
else:
    from importlib_metadata import Distribution, distribution

import gptsum
import gptsum.checksum
import gptsum.gpt

__metadata__ = cast(Callable[[str], Distribution], distribution)(__name__).metadata
__author__ = __metadata__["Author"]
"""Package author name.

.. versionadded:: 0.0.9
"""
__contact__ = __metadata__["Author-email"]
"""Package author contact e-mail address.

.. versionadded:: 0.0.9
"""
__license__ = __metadata__["License"]
"""Package license identifier.

.. versionadded:: 0.0.9
"""
__version__ = __metadata__["Version"]
"""Package version identifier.

.. versionadded:: 0.0.1
"""
del __metadata__


def _check_fd_or_path(fd: Optional[int], path: Optional[Path]) -> None:
    """Ensure either ``fd`` or ``path`` is given.

    >>> _check_fd_or_path(None, None)
    Traceback (most recent call last):
    ...
    ValueError: Either fd or path must be given

    >>> _check_fd_or_path(0, Path("/"))
    Traceback (most recent call last):
    ...
    ValueError: Both fd and path can't be given

    :param fd: Possible non-``None`` FD value
    :param path: Possiblel non-``None`` path value

    :raises ValueError: Both ``fd`` and ``path`` are ``None``
    :raises ValueError: Both ``fd`` and ``path`` are not ``None``
    """
    if fd is None and path is None:
        raise ValueError("Either fd or path must be given")
    if fd is not None and path is not None:
        raise ValueError("Both fd and path can't be given")


def get_guid(*, fd: Optional[int] = None, path: Optional[Path] = None) -> uuid.UUID:
    """Get the GUID of a disk image.

    One of ``fd`` or ``path`` must be given.

    :param fd: Readable file-descriptor to an image file
    :param path: Path of a readable image file

    :returns: GUID of the disk image

    .. versionadded:: 0.0.9
    """
    _check_fd_or_path(fd, path)

    with gptsum.gpt.GPTImage(fd=fd, path=path, open_mode=os.O_RDONLY) as image:
        image.validate()
        primary = image.read_primary_gpt_header()
        return primary.disk_guid


def set_guid(
    guid: uuid.UUID, *, fd: Optional[int] = None, path: Optional[Path] = None
) -> None:
    """Set the GUID of a disk image.

    One of ``fd`` or ``path`` must be given.

    :param guid: GUID to write to the disk image
    :param fd: Readable and writable file-descriptor to an image file
    :param path: Path of a readable and writable image file

    .. versionadded:: 0.0.9
    """
    _check_fd_or_path(fd, path)

    with gptsum.gpt.GPTImage(fd=fd, path=path, open_mode=os.O_RDWR) as image:
        image.validate()
        image.update_guid(guid)


def calculate_expected_guid(
    *, fd: Optional[int] = None, path: Optional[Path] = None
) -> uuid.UUID:
    """Calculate the expected checksum GUID of a disk image.

    One of ``fd`` or ``path`` must be given.

    :param fd: Readable file-descriptor to an image file
    :param path: Path of a readable image file

    :returns: Expected checksum GUID of the disk image

    .. versionadded:: 0.0.9
    """
    _check_fd_or_path(fd, path)

    with gptsum.gpt.GPTImage(fd=fd, path=path, open_mode=os.O_RDONLY) as image:
        image.validate()
        digest = gptsum.checksum.calculate(image)
        return gptsum.checksum.digest_to_guid(digest)


def embed(*, fd: Optional[int] = None, path: Optional[Path] = None) -> None:
    """Embed the calculated checksum GUID in an image file.

    In essence a combination of :func:`calculate_expected_guid` and :func:`set_guid`.

    One of ``fd`` or ``path`` must be given.

    :param fd: Readable and writable file-descriptor to an image file
    :param path: Path to a readable and writable image file

    .. versionadded:: 0.0.9
    """
    _check_fd_or_path(fd, path)

    current_guid = get_guid(fd=fd, path=path)
    checksum_guid = calculate_expected_guid(fd=fd, path=path)

    if current_guid != checksum_guid:
        set_guid(checksum_guid, fd=fd, path=path)


@dataclasses.dataclass(frozen=True)
class VerificationFailure(Exception):
    """Exception raised when :func:`verify` fails.

    .. versionadded:: 0.0.9
    """

    expected: uuid.UUID
    actual: uuid.UUID


def verify(*, fd: Optional[int] = None, path: Optional[Path] = None) -> None:
    """Verify a GPT disk image GUID against the calculated checksum.

    In essence a combination of :func:`calculate_expected_guid` and :func:`get_guid`.

    One of ``fd`` or ``path`` must be given.

    :param fd: Readable file-descriptor to an image file
    :param path: Path to a readable image file

    :raises VerificationFailure: Current GUID and checksum mismatch

    .. versionadded:: 0.0.9
    """
    _check_fd_or_path(fd, path)

    current_guid = get_guid(fd=fd, path=path)
    checksum_guid = calculate_expected_guid(fd=fd, path=path)

    if current_guid != checksum_guid:
        raise VerificationFailure(checksum_guid, current_guid)



"""Fake test module so :mod:`gptsum.__main__` gets imported."""

import runpy
import sys

from pytest_mock import MockerFixture


def test_main(mocker: MockerFixture) -> None:
    """Test running the :mod:`gptsum` package."""
    mocker.patch("sys.argv", sys.argv[:1])
    # Once for real
    runpy.run_module("gptsum", run_name="__main__", alter_sys=True)
    # Once to please the coverage checks
    runpy.run_module("gptsum", alter_sys=True)



"""Tests for the :mod:`gptsum.cli` module."""

import hashlib
import os
import uuid
from pathlib import Path

import pytest

import gptsum
from gptsum import checksum, cli
from tests import conftest


def test_no_arguments(capsys: pytest.CaptureFixture[str]) -> None:
    """Test the CLI without any options."""
    cli.main([])

    captured = capsys.readouterr()
    assert captured.out.startswith("usage: ")


def test_version(capsys: pytest.CaptureFixture[str]) -> None:
    """Test the CLI :option:`--verbose` option."""
    with pytest.raises(SystemExit):
        cli.main(["--version"])

    captured = capsys.readouterr()
    assert captured.out == "{}\n".format(gptsum.__version__)


@pytest.mark.parametrize(
    ("disk_file", "expected_guid"),
    [
        (conftest.TESTDATA_DISK, conftest.TESTDATA_DISK_GUID),
        (conftest.TESTDATA_EMBEDDED_DISK, conftest.TESTDATA_EMBEDDED_DISK_GUID),
    ],
)
def test_get_guid(
    capsys: pytest.CaptureFixture[str],
    disk_file: Path,
    expected_guid: uuid.UUID,
) -> None:
    """Test the CLI :option:`get-guid` subcommand."""
    cli.main(["get-guid", str(disk_file)])

    captured = capsys.readouterr()
    assert captured.out == "{}\n".format(expected_guid)


@pytest.mark.parametrize(
    ("disk_file", "expected_guid"),
    [
        (conftest.TESTDATA_DISK, conftest.TESTDATA_EMBEDDED_DISK_GUID),
        (conftest.TESTDATA_EMBEDDED_DISK, conftest.TESTDATA_EMBEDDED_DISK_GUID),
    ],
)
def test_calculate_expected_guid(
    capsys: pytest.CaptureFixture[str], disk_file: Path, expected_guid: uuid.UUID
) -> None:
    """Test the CLI :option:`calculate-expected-guid` subcommand."""
    cli.main(["calculate-expected-guid", str(disk_file)])

    captured = capsys.readouterr()
    assert captured.out == "{}\n".format(expected_guid)


def test_set_guid(capsys: pytest.CaptureFixture[str], disk_image: Path) -> None:
    """Test the CLI :option:`set-guid` subcommand."""
    new_guid = uuid.UUID("bf85e1a8-748e-46f8-909f-4a70067efbe2")

    cli.main(["set-guid", str(disk_image), str(new_guid)])

    cli.main(["get-guid", str(disk_image)])

    captured = capsys.readouterr()
    assert captured.out == "{}\n".format(new_guid)


def test_embed(capsys: pytest.CaptureFixture[str], disk_image: Path) -> None:
    """Test the CLI :option:`embed` subcommand."""
    cli.main(["embed", str(disk_image)])

    cli.main(["get-guid", str(disk_image)])

    captured = capsys.readouterr()
    assert captured.out == "{}\n".format(conftest.TESTDATA_EMBEDDED_DISK_GUID)

    hash1 = hashlib.sha256()
    with open(conftest.TESTDATA_EMBEDDED_DISK, "rb") as fd:
        size = os.fstat(fd.fileno()).st_size
        done = checksum.hash_file(hash1.update, fd.fileno(), size, 0)
        assert done == size

    hash2 = hashlib.sha256()
    with open(disk_image, "rb") as fd:
        size = os.fstat(fd.fileno()).st_size
        done = checksum.hash_file(hash2.update, fd.fileno(), size, 0)
        assert done == size

    assert hash1.hexdigest() == hash2.hexdigest()


def test_embed_noop(disk_image: Path) -> None:
    """Test to ensure issuing :option:`embed` on an already-prepared file is a no-op."""
    cli.main(["embed", str(disk_image)])
    stat1 = os.stat(disk_image)
    cli.main(["embed", str(disk_image)])
    stat2 = os.stat(disk_image)

    assert stat2.st_mtime == stat1.st_mtime
    assert stat2.st_ctime == stat1.st_ctime


def test_verify(disk_image: Path) -> None:
    """Test the CLI :option:`verify` subcommand."""
    cli.main(["verify", str(conftest.TESTDATA_EMBEDDED_DISK)])

    with pytest.raises(SystemExit):
        cli.main(["verify", str(disk_image)])

    cli.main(["embed", str(disk_image)])
    cli.main(["verify", str(disk_image)])
